#!/usr/bin/env python2
# -*- coding: UTF-8 -*-

"""
PT2/SS18/U3/Validator/v3

Danke, dass du dieses Validatorscript heruntergeladen hast.
Wenn es dir lokal noch hilfreicher sein soll, kannst du dir
die Fehler rot einfärben lassen.

Bash-Nutzer führen dazu einfach folgendes aus (alle anderen googeln):

	$ color() (set -o pipefail; "$@" 2>&1>&3 | sed $'s,.*,\e[31m&\e[m,' >&2) 3>&1
	$ color python2 path/to/validator.py

Viel Spaß!
"""

import operator
import os
import subprocess
import sys

#-------------------------------------------------------------------------------

def log(msg="", newLine=True, okay=False):
	f = sys.stdout if okay else sys.stderr
	f.write("%s\n" % msg if newLine else msg)
	f.flush()

def silent_atoi(s):
	try:
		return int(s)
	except ValueError:
		return None

#-------------------------------------------------------------------------------

class Test(object):

	def exec_(self, input_text, *cmd_and_args):
		p = subprocess.Popen(cmd_and_args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		return tuple(map(lambda s: s.strip(), p.communicate(input=input_text)[:2])) + (p.returncode, )

	def run_prog(self, *cmd_and_args):
		return self.exec_("", *cmd_and_args)[0]

	def test(self):
		self.run()
		okay = self.okay()
		log(self.what(), okay=okay)
		return okay

	def run(self):
		raise NotImplementedError

	def what(self):
		raise NotImplementedError

	def okay(self):
		raise NotImplementedError

class EmptyTestLine(Test):

	def __init__(self, text=""):
		self.text = text

	def run(self):
		pass

	def what(self):
		return self.text

	def okay(self):
		return True

class TestGroup(Test):
	"""A group of tests, where all tests are executed"""

	def __init__(self, *tests):
		self.tests = tests
		self.all_okay = False

	def test(self):
		self.all_okay = all([t.test() for t in self.tests])
		return self.all_okay

	def okay(self):
		return self.all_okay

class AbortingTestGroup(TestGroup):
	"""A group of tests, where the first failed test stops the entire group"""

	def test(self):
		self.all_okay = all(t.test() for t in self.tests)
		return self.all_okay

class ReturnCodeTest(Test):

	def __init__(self, progname, args=None, retcode=0, showstdout=False, showstderr=False, inputText=""):
		self.progname = progname
		self.inputText = inputText
		self.cmdline = [progname] + map(str, args or [])
		self.expected = retcode
		self.exception = None
		self.showstdout = showstdout
		self.showstderr = showstderr
		self.output = None
		self.stdout = ""
		self.stderr = ""

	def _get_output(self, tpl):
		self.stdout = tpl[0]
		self.stderr = tpl[1]
		return tpl[2]

	def run(self):
		try:
			self.output = self._get_output(self.exec_(self.inputText, *self.cmdline))
		except Exception as e:
			self.exception = e
	
	def _s(self, a):
		return " ".join(map(str, a))

	def _formatCall(self):
		if self.inputText:
			return "echo -ne %r | %s" % (self.inputText, self._s(self.cmdline))
		return self._s(self.cmdline)

	def what(self):
		return "$ %s #- exited with %s (expected %d)%s%s" % (
					self._formatCall(),
					str(self.output) if self.exception is None else repr(str(self.exception)),
					self.expected,
					"" if not (self.showstdout and self.stdout) else ("\n%s" % self.stdout),
					"" if not (self.showstderr and self.stderr) else ("\n%s" % self.stderr),
				)

	def okay(self):
		return self.output == self.expected

class ExecutionTest(ReturnCodeTest):
	"""Run a programm and compare the lines of stdout to a list of expected output."""

	def __init__(self, progname, args, expected, compare=operator.eq, key=lambda x: x, inputText = "", showstderr=False):
		super(ExecutionTest, self).__init__(progname, args, expected, inputText=inputText, showstderr=showstderr)
		self.compare = compare
		self.key = key

	def _get_output(self, tpl):
		super(ExecutionTest, self)._get_output(tpl)
		return self.stdout.split("\n")

	def what(self):
		_r = repr
		return "$ " + self._formatCall() + "\n" + \
				("" if not (self.showstderr and self.stderr) else ("%s\n" % self.stderr)) + \
				(("received: " + _r(self.output))
						if self.exception is None
						else str(self.exception)) + "\n" + \
			   "expected: " + _r(self.expected)

	def okay(self):
		return self.exception is None and self.compare(self.key(self.output), self.key(self.expected))

#-------------------------------------------------------------------------------

def validate():
	tests = TestGroup(
			AbortingTestGroup(
				ExecutionTest("java", ["Minimum"], ["2"], inputText="8 7 2 3 4 9 -1"),
				ExecutionTest("java", ["Minimum"], ["4"], inputText="42 1337 21 4 5 -2 3"),
				ExecutionTest("java", ["Minimum"], ["" ], inputText="-3"),
			),
			AbortingTestGroup(
				ExecutionTest("java", ["C2F", "0", "100"], ["32.0", "212.0"], showstderr=True),
				ExecutionTest("java", ["C2F", "-20", "13.37", "4211"], ["-4.0", "56.066", "7611.8"], showstderr=True)
			),
			AbortingTestGroup(
				ExecutionTest("java", ["F2C", "32", "212"], ["0.0", "100.0"]),
				ExecutionTest("java", ["F2C", "-22", "50", "1337"], ["-30.0", "10.0", "725.0"])
			),
			ReturnCodeTest("java", ["-ea", "AccountTest"], showstdout=True, showstderr=True),
	)

	return tests.test()

if __name__ == "__main__":
	sys.exit(0 if validate() else 1)
 
