#!/usr/bin/env python2
# -*- coding: UTF-8 -*-

"""
PT2/SS18/U5/Validator/v2

Danke, dass du dieses Validatorscript heruntergeladen hast.
Wenn es dir lokal noch hilfreicher sein soll, kannst du dir
die Fehler rot einfärben lassen.

Bash-Nutzer führen dazu einfach folgendes aus (alle anderen googeln):

	$ color() (set -o pipefail; "$@" 2>&1>&3 | sed $'s,.*,\e[31m&\e[m,' >&2) 3>&1
	$ color python2 path/to/validator.py

Viel Spaß!
"""

import operator
import os
import subprocess
import sys

#-------------------------------------------------------------------------------

def log(msg="", newLine=True, okay=False):
	f = sys.stdout if okay else sys.stderr
	f.write("%s\n" % msg if newLine else msg)
	f.flush()

def silent_atoi(s):
	try:
		return int(s)
	except ValueError:
		return None

#-------------------------------------------------------------------------------

class Test(object):

	def exec_(self, input_text, *cmd_and_args):
		p = subprocess.Popen(cmd_and_args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		return tuple(map(lambda s: s.strip(), p.communicate(input=input_text)[:2])) + (p.returncode, )

	def run_prog(self, *cmd_and_args):
		return self.exec_("", *cmd_and_args)[0]

	def test(self):
		self.run()
		okay = self.okay()
		log(self.what(), okay=okay)
		return okay

	def run(self):
		raise NotImplementedError

	def what(self):
		raise NotImplementedError

	def okay(self):
		raise NotImplementedError

class EmptyTestLine(Test):

	def __init__(self, text=""):
		self.text = text

	def run(self):
		pass

	def what(self):
		return self.text

	def okay(self):
		return True

class TestGroup(Test):
	"""A group of tests, where all tests are executed"""

	def __init__(self, *tests):
		self.tests = tests
		self.all_okay = False

	def test(self):
		self.all_okay = all([t.test() for t in self.tests])
		return self.all_okay

	def okay(self):
		return self.all_okay

class AbortingTestGroup(TestGroup):
	"""A group of tests, where the first failed test stops the entire group"""

	def test(self):
		self.all_okay = all(t.test() for t in self.tests)
		return self.all_okay

class FileMimeTest(Test):
	"""Is the specified file existant of a given mimetype?"""

	def __init__(self, fname, expected_mime_toplevel, expected_mime_leaf="*"):
		self.fname = fname
		self.expected_mime_toplevel = expected_mime_toplevel
		self.expected_mime_leaf = expected_mime_leaf
		self.ttype = None
		self.ltype = None

	def run(self):
		self.exists = os.path.exists(self.fname)
		if self.exists:
			self.mimetype = self.run_prog("file", "--mime-type", "-b", self.fname)
			self.ttype, self.ltype = self.mimetype.split("/", 1)

	def what(self):
		if not self.exists:
			return '# file "%s" does not exist' % self.fname
		return '# file "%s" exists and is "%s" (expected "%s/%s")' % (self.fname, self.mimetype, self.expected_mime_toplevel, self.expected_mime_leaf)

	def okay(self):
		leaf_matches = (self.expected_mime_leaf == "*") or (self.ltype == self.expected_mime_leaf)
		return self.exists and self.ttype == self.expected_mime_toplevel and leaf_matches

class ReturnCodeTest(Test):

	def __init__(self, progname, args=None, retcode=0, showstdout=False, showstderr=False, inputText=""):
		self.progname = progname
		self.inputText = inputText
		self.cmdline = [progname] + map(str, args or [])
		self.expected = retcode
		self.exception = None
		self.showstdout = showstdout
		self.showstderr = showstderr
		self.output = None
		self.stdout = ""
		self.stderr = ""

	def _get_output(self, tpl):
		self.stdout = tpl[0]
		self.stderr = tpl[1]
		return tpl[2]

	def run(self):
		try:
			self.output = self._get_output(self.exec_(self.inputText, *self.cmdline))
		except Exception as e:
			self.exception = e
	
	def _s(self, a):
		return " ".join(map(str, a))

	def _formatCall(self):
		if self.inputText:
			return "echo -ne %r | %s" % (self.inputText, self._s(self.cmdline))
		return self._s(self.cmdline)

	def what(self):
		return "$ %s #- exited with %s (expected %d)%s%s" % (
					self._formatCall(),
					str(self.output) if self.exception is None else repr(str(self.exception)),
					self.expected,
					"" if not (self.showstdout and self.stdout) else ("\n%s" % self.stdout),
					"" if not (self.showstderr and self.stderr) else ("\n%s" % self.stderr),
				)

	def okay(self):
		return self.output == self.expected

#-------------------------------------------------------------------------------

class PrintFileIfExists(Test):

	def __init__(self, fname=""):
		self.fname = fname
		self.text = ""

	def run(self):
		try:
			with open(self.fname) as f:
				self.text = f.read()
		except:
			pass

	def what(self):
		return self.text

	def okay(self):
		return True

#-------------------------------------------------------------------------------

def validate():
	tests = AbortingTestGroup(
		FileMimeTest("performance_tests.txt", "text"),
		TestGroup(
			ReturnCodeTest("ant", ["test"], showstdout=True, showstderr=True),
			PrintFileIfExists("TEST-de.uni_potsdam.hpi.ArrayDequeTest.txt"),
			PrintFileIfExists("TEST-de.uni_potsdam.hpi.LinkedDequeTest.txt"),
		),
		ReturnCodeTest("java", ["-cp", "build", "DequePerformance"], showstdout=True, showstderr=True),
	)

	return tests.test()

if __name__ == "__main__":
	sys.exit(0 if validate() else 1)
 
