#!/usr/bin/env python2
# -*- coding: UTF-8 -*-

"""
PT2/SS18/U6/Validator/v5

Danke, dass du dieses Validatorscript heruntergeladen hast.
Wenn es dir lokal noch hilfreicher sein soll, kannst du dir
die Fehler rot einfärben lassen.

Bash-Nutzer führen dazu einfach folgendes aus (alle anderen googeln):

	$ color() (set -o pipefail; "$@" 2>&1>&3 | sed $'s,.*,\e[31m&\e[m,' >&2) 3>&1
	$ color python2 path/to/validator.py

Viel Spaß!
"""

import operator
import os
import subprocess
import sys

#-------------------------------------------------------------------------------

def log(msg="", newLine=True, okay=False):
	f = sys.stdout if okay else sys.stderr
	f.write("%s\n" % msg if newLine else msg)
	f.flush()

def silent_atoi(s):
	try:
		return int(s)
	except ValueError:
		return None

#-------------------------------------------------------------------------------

class Test(object):

	def exec_(self, input_text, *cmd_and_args):
		p = subprocess.Popen(cmd_and_args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		return tuple(map(lambda s: s.strip(), p.communicate(input=input_text)[:2])) + (p.returncode, )

	def run_prog(self, *cmd_and_args):
		return self.exec_("", *cmd_and_args)[0]

	def test(self):
		self.run()
		okay = self.okay()
		log(self.what(), okay=okay)
		return okay

	def run(self):
		raise NotImplementedError

	def what(self):
		raise NotImplementedError

	def okay(self):
		raise NotImplementedError


class EmptyTestLine(Test):

	def __init__(self, text=""):
		self.text = text

	def run(self):
		pass

	def what(self):
		return self.text

	def okay(self):
		return True


class TestGroup(Test):
	"""A group of tests, where all tests are executed"""

	def __init__(self, *tests):
		self.tests = tests
		self.is_okay = False

	def test(self):
		self.is_okay = all([t.test() for t in self.tests])
		return self.is_okay

	def okay(self):
		return self.is_okay


class AbortingTestGroup(TestGroup):
	"""A group of tests, where the first failed test stops the entire group"""

	def test(self):
		self.is_okay = all(t.test() for t in self.tests)
		return self.is_okay


class AlternativeTestGroup(TestGroup):
	"""A group of tests, where at least one test has to pass"""

	def test(self):
		self.is_okay = any(t.test() for t in self.tests)
		return self.is_okay


class FileMimeTest(Test):
	"""Is the specified file existant of a given mimetype?"""

	def __init__(self, fname, expected_mime_toplevel, expected_mime_leaf="*"):
		self.fname = fname
		self.expected_mime_toplevel = expected_mime_toplevel
		self.expected_mime_leaf = expected_mime_leaf
		self.ttype = None
		self.ltype = None

	def run(self):
		self.exists = os.path.exists(self.fname)
		if self.exists:
			self.mimetype = self.run_prog("file", "--mime-type", "-b", self.fname)
			self.ttype, self.ltype = self.mimetype.split("/", 1)

	def what(self):
		if not self.exists:
			return '# file "%s" does not exist' % self.fname
		return '# file "%s" exists and is "%s" (expected "%s/%s")' % (self.fname, self.mimetype, self.expected_mime_toplevel, self.expected_mime_leaf)

	def okay(self):
		leaf_matches = (self.expected_mime_leaf == "*") or (self.ltype == self.expected_mime_leaf)
		return self.exists and self.ttype == self.expected_mime_toplevel and leaf_matches


class ReturnCodeTest(Test):

	def __init__(self, progname, args=None, retcode=0, showstdout=False, showstderr=False, inputText="", truncateLines=None):
		self.progname = progname
		self.inputText = inputText
		self.cmdline = [progname] + map(str, args or [])
		self.expected = retcode
		self.exception = None
		self.showstdout = showstdout
		self.showstderr = showstderr
		self.output = None
		self.truncateLines = truncateLines
		self.stdout = ""
		self.stderr = ""

	def _truncate(self, text):
		# Hoping for PEP 572 to fix this mess
		#    https://www.python.org/dev/peps/pep-0572/
		# Okay, hoping for submit-exec to support Python 3.8

		lines = text.split("\n")
		if self.truncateLines is None or len(lines) <= self.truncateLines:
			return text

		return "\n".join(lines[:self.truncateLines] + ["", "[OUTPUT TRUNCATED]"])

	def _get_output(self, tpl):
		self.stdout = tpl[0]
		self.stderr = tpl[1]
		return tpl[2]

	def run(self):
		try:
			self.output = self._get_output(self.exec_(self.inputText, *self.cmdline))
		except Exception as e:
			self.exception = e

	def _s(self, a):
		return " ".join(map(str, a))

	def _formatCall(self):
		if self.inputText:
			return "echo -ne %r | %s" % (self.inputText, self._s(self.cmdline))
		return self._s(self.cmdline)

	def what(self):
		return "$ %s #- exited with %s (expected %d)%s%s" % (
					self._formatCall(),
					str(self.output) if self.exception is None else repr(str(self.exception)),
					self.expected,
					"" if not (self.showstdout and self.stdout) else ("\n%s" % self._truncate(self.stdout)),
					"" if not (self.showstderr and self.stderr) else ("\n%s" % self._truncate(self.stderr)),
				)

	def okay(self):
		return self.output == self.expected


class ExecutionTest(ReturnCodeTest):
	"""Run a programm and compare the lines of stdout to a list of expected output."""

	def __init__(self, progname, args, expected, compare=operator.eq, key=lambda x: x, inputText = "", showstderr=False):
		super(ExecutionTest, self).__init__(progname, args, expected, inputText=inputText, showstderr=showstderr)
		self.compare = compare
		self.key = key

	def _get_output(self, tpl):
		super(ExecutionTest, self)._get_output(tpl)
		return self.stdout.split("\n")

	def what(self):
		_r = repr
		return "$ " + self._formatCall() + "\n" + \
				("" if not (self.showstderr and self.stderr) else ("%s\n" % self.stderr)) + \
				(("received: " + _r(self.output))
						if self.exception is None
						else str(self.exception)) + "\n" + \
			   "expected: " + _r(self.expected)

	def okay(self):
		return self.exception is None and self.compare(self.key(self.output), self.key(self.expected))

#-------------------------------------------------------------------------------

import glob
import re

class ReviewFinder(TestGroup):

	def __init__(self):
		self.names = [match.group(1) for match in
		                 (re.match(r"review_(.*).txt", n) for n in glob.glob("review_*.txt"))
				      if match]
		super(ReviewFinder, self).__init__(*(
		                 FileMimeTest("review_%s.txt" % name, "text") for name in self.names
			          ))
		log("Found %d review(s): %s" % (len(self.names), ", ".join(self.names)), okay=self.at_least_one_review())

	def at_least_one_review(self):
		return len(self.names) >= 1

	def test(self):
		return self.at_least_one_review() and super(ReviewFinder, self).test()

	def okay(self):
		return self.at_least_one_review() and super(ReviewFinder, self).okay()

#-------------------------------------------------------------------------------

def validate():
	_w_classp = lambda *l: ("-cp", "build") + l
	tests = TestGroup(
		ReviewFinder(),
		EmptyTestLine("# Trying SortRandom as test runner and as test case. Need only one to succeed."),
		AlternativeTestGroup(
			ReturnCodeTest("java", _w_classp("SortRandom", "-ea"), showstdout=True, showstderr=True, truncateLines=48),
			ReturnCodeTest("ant", ["test"], showstdout=True, showstderr=True, truncateLines=48),
		),
		AbortingTestGroup(
			ExecutionTest("java", _w_classp("SortStdin"), ["Apples", "Mangos", "Oranges", "Watermelons"],
			                          inputText="\n".join(["Oranges", "Watermelons", "Mangos", "Apples"]),
			                          showstderr=True),
			ExecutionTest("java", _w_classp("SortStdin"), ["1", "11", "12", "2", "22", "3"],
			                          inputText="\n".join(["11", "3", "1", "22", "2", "12"])),
			ExecutionTest("java", _w_classp("SortStdin"), list("AAEEGILMNOPRSTX"), inputText="\n".join("ASORTINGEXAMPLE")),
			ExecutionTest("java", _w_classp("SortStdin"), [""], inputText=""),
			AbortingTestGroup(
				ExecutionTest("java", _w_classp("SortStdin", "-c", "1"), ["1 A b", "2 A c", "2 A a", "3 A a"],
				                                     inputText="\n".join(["2 A c", "3 A a", "1 A b", "2 A a"]),
			                                         showstderr=True),
				ExecutionTest("java", _w_classp("SortStdin", "-c", "2"), ["1 A b", "2 A c", "2 A a", "3 A a"],
				                                     inputText="\n".join(["1 A b", "2 A c", "2 A a", "3 A a"])),
				ExecutionTest("java", _w_classp("SortStdin", "-c", "3"), ["3 A a", "2 A a", "1 A b", "2 A c"],
				                                     inputText="\n".join(["2 A c", "3 A a", "1 A b", "2 A a"])),
				ExecutionTest("java", _w_classp("SortStdin", "-c", "3"), ["2   A  a", "3\t A  a", "1    A b", "2 A\tc"],
				                                     inputText="\n".join(["1    A b", "2 A\tc", "2   A  a", "3\t A  a"])),
			),
		),
	)

	return tests.test()

if __name__ == "__main__":
	sys.exit(0 if validate() else 1)

