#!/usr/bin/env python2
# -*- coding: UTF-8 -*-

"""
PT2/SS18/U7/Validator/v1

Danke, dass du dieses Validatorscript heruntergeladen hast.
Wenn es dir lokal noch hilfreicher sein soll, kannst du dir
die Fehler rot einfärben lassen.

Bash-Nutzer führen dazu einfach folgendes aus (alle anderen googeln):

	$ color() (set -o pipefail; "$@" 2>&1>&3 | sed $'s,.*,\e[31m&\e[m,' >&2) 3>&1
	$ color python2 path/to/validator.py

Viel Spaß!
"""

import operator
import os
import subprocess
import sys

#-------------------------------------------------------------------------------

def log(msg="", newLine=True, okay=False):
	f = sys.stdout if okay else sys.stderr
	f.write("%s\n" % msg if newLine else msg)
	f.flush()

def silent_atoi(s):
	try:
		return int(s)
	except ValueError:
		return None

#-------------------------------------------------------------------------------

class Test(object):

	def exec_(self, input_text, *cmd_and_args):
		p = subprocess.Popen(cmd_and_args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		return tuple(map(lambda s: s.strip(), p.communicate(input=input_text)[:2])) + (p.returncode, )

	def run_prog(self, *cmd_and_args):
		return self.exec_("", *cmd_and_args)[0]

	def test(self):
		self.run()
		okay = self.okay()
		log(self.what(), okay=okay)
		return okay

	def run(self):
		raise NotImplementedError

	def what(self):
		raise NotImplementedError

	def okay(self):
		raise NotImplementedError


class TestGroup(Test):
	"""A group of tests, where all tests are executed"""

	def __init__(self, *tests):
		self.tests = tests
		self.is_okay = False

	def test(self):
		self.is_okay = all([t.test() for t in self.tests])
		return self.is_okay

	def okay(self):
		return self.is_okay


class AbortingTestGroup(TestGroup):
	"""A group of tests, where the first failed test stops the entire group"""

	def test(self):
		self.is_okay = all(t.test() for t in self.tests)
		return self.is_okay


class ReturnCodeTest(Test):

	def __init__(self, progname, args=None, retcode=0, showstdout=False, showstderr=False, inputText="", truncateLines=None):
		self.progname = progname
		self.inputText = inputText
		self.cmdline = [progname] + map(str, args or [])
		self.expected = retcode
		self.exception = None
		self.showstdout = showstdout
		self.showstderr = showstderr
		self.output = None
		self.truncateLines = truncateLines
		self.stdout = ""
		self.stderr = ""

	def _truncate(self, text):
		# Hoping for PEP 572 to fix this mess
		#    https://www.python.org/dev/peps/pep-0572/
		# Okay, hoping for submit-exec to support Python 3.8

		lines = text.split("\n")
		if self.truncateLines is None or len(lines) <= self.truncateLines:
			return text

		return "\n".join(lines[:self.truncateLines] + ["", "[OUTPUT TRUNCATED]"])

	def _get_output(self, tpl):
		self.stdout = tpl[0]
		self.stderr = tpl[1]
		return tpl[2]

	def run(self):
		try:
			self.output = self._get_output(self.exec_(self.inputText, *self.cmdline))
		except Exception as e:
			self.exception = e

	def _s(self, a):
		return " ".join(map(str, a))

	def _formatCall(self):
		if self.inputText:
			return "echo -ne %r | %s" % (self.inputText, self._s(self.cmdline))
		return self._s(self.cmdline)

	def what(self):
		return "$ %s #- exited with %s (expected %d)%s%s" % (
					self._formatCall(),
					str(self.output) if self.exception is None else repr(str(self.exception)),
					self.expected,
					"" if not (self.showstdout and self.stdout) else ("\n%s" % self._truncate(self.stdout)),
					"" if not (self.showstderr and self.stderr) else ("\n%s" % self._truncate(self.stderr)),
				)

	def okay(self):
		return self.output == self.expected


class ExecutionTest(ReturnCodeTest):
	"""Run a programm and compare the lines of stdout to a list of expected output."""

	def __init__(self, progname, args, expected, compare=operator.eq, key=lambda x: x, inputText = "", showstderr=False):
		super(ExecutionTest, self).__init__(progname, args, expected, inputText=inputText, showstderr=showstderr)
		self.compare = compare
		self.key = key

	def _get_output(self, tpl):
		super(ExecutionTest, self)._get_output(tpl)
		return self.stdout.split("\n")

	def what(self):
		_r = repr
		return "$ " + self._formatCall() + "\n" + \
				("" if not (self.showstderr and self.stderr) else ("%s\n" % self.stderr)) + \
				(("received: " + _r(self.output))
						if self.exception is None
						else str(self.exception)) + "\n" + \
			   "expected: " + _r(self.expected)

	def okay(self):
		return self.exception is None and self.compare(self.key(self.output), self.key(self.expected))

#-------------------------------------------------------------------------------

def SWIPLTest(filename, *args, **kwargs):
	return ExecutionTest("swipl", ["-q", filename], *args, **kwargs)

#-------------------------------------------------------------------------------

def validate():
	tests = TestGroup(
		AbortingTestGroup(
			SWIPLTest("indexfilter.pl", ["Y = [b, d, f, h]."], showstderr=True,
		                       inputText="odd_indices([a, b, c, d, e, f, g, h], Y)."),
			SWIPLTest("indexfilter.pl", ["Y = [b, d, f, h]."], showstderr=True,
		                       inputText="odd_indices([a, b, c, d, e, f, g, h, i], Y),!."),
			SWIPLTest("indexfilter.pl", ["Y = [b, d, f, h]."], showstderr=True,
		                       inputText="odd_indices([a, b, c, d, e, f, g, h, i], Y)."),
			SWIPLTest("indexfilter.pl", ["Y = []."], showstderr=True,
		                       inputText="odd_indices([], Y)."),
			SWIPLTest("indexfilter.pl", ["true."], showstderr=True,
		                       inputText="odd_indices([a, b, c], [b])."),
			SWIPLTest("indexfilter.pl", ["false."], showstderr=True,
		                       inputText="odd_indices([a, b, c], [c])."),
			SWIPLTest("indexfilter.pl", ["X = b,", "Y = d."], showstderr=True,
		                       inputText="odd_indices([a, X, c, Y], [b, d])."),
		),
		AbortingTestGroup(
			SWIPLTest("sort.pl",        ["L = [3, 4, 5, 5, 7]."], showstderr=True,
		                       inputText="quicksort([7, 4, 5, 3, 5], L),!."),
			SWIPLTest("sort.pl",        ["L = [3, 4, 5, 5, 7, 8]."], showstderr=True,
		                       inputText="quicksort([7, 4, 5, 8, 3, 5], L),!."),
			SWIPLTest("sort.pl",        ["L = [3, 4, 5, 5, 7]."], showstderr=True,
		                       inputText="mergesort([7, 4, 5, 3, 5], L),!."),
			SWIPLTest("sort.pl",        ["L = [3, 4, 5, 5, 7, 8]."], showstderr=True,
		                       inputText="mergesort([7, 4, 5, 8, 3, 5], L),!."),
		),
		SWIPLTest("raetsel.pl", 
		                       ['david ist 11 Jahre alt, wohnt in berlin und spielt fussball.',
								'alexander ist 12 Jahre alt, wohnt in leipzig und spielt schach.',
								'lukas ist 13 Jahre alt, wohnt in hamburg und spielt handball.',
								'florian ist 14 Jahre alt, wohnt in mainz und spielt klavier.'],
							   showstderr=True, inputText="halt."),
	)

	return tests.test()

if __name__ == "__main__":
	sys.exit(0 if validate() else 1)

